@@ -3,8 +3,8 @@
 '''
 A library to access Tryton's models like a client.
 '''
-__version__ = "3.2.2"
-__all__ = ['Model', 'Wizard']
+__version__ = "3.4.1"
+__all__ = ['Model', 'Wizard', 'Report']
 import sys
 try:
     import cdecimal
@@ -26,23 +26,32 @@ _MODELS = threading.local()
 
 class _EvalEnvironment(dict):
     'Dictionary for evaluation'
-    def __init__(self, parent):
+    def __init__(self, parent, eval_type='eval'):
         super(_EvalEnvironment, self).__init__()
         self.parent = parent
+        assert eval_type in ('eval', 'on_change')
+        self.eval_type = eval_type
 
     def __getitem__(self, item):
         if item == '_parent_' + self.parent._parent_name \
-                and self.parent.parent:
-            return _EvalEnvironment(self.parent.parent)
-        return self.parent._get_eval()[item]
+                and self.parent._parent:
+            return _EvalEnvironment(self.parent._parent,
+                eval_type=self.eval_type)
+        if self.eval_type == 'eval':
+            return self.parent._get_eval()[item]
+        else:
+            return self.parent._get_on_change_values(fields=[item])[item]
 
     def __getattr__(self, item):
-        return self.__getitem__(item)
+        try:
+            return self.__getitem__(item)
+        except KeyError:
+            raise AttributeError(item)
 
     def get(self, item, default=None):
         try:
-            return self.__getattr__(item)
-        except:
+            return self.__getitem__(item)
+        except KeyError:
             pass
         return super(_EvalEnvironment, self).get(item, default)
 
@@ -55,7 +64,13 @@ class _EvalEnvironment(dict):
     __repr__ = __str__
 
     def __contains__(self, item):
-        return item in self.parent._fields
+        if item == '_parent_' + self.parent._parent_name \
+                and self.parent._parent:
+            return True
+        if self.eval_type == 'eval':
+            return item in self.parent._get_eval()
+        else:
+            return item in self.parent._fields
 
 
 class FieldDescriptor(object):
@@ -97,23 +112,16 @@ class CharDescriptor(FieldDescriptor):
     default = None
 
     def __set__(self, instance, value):
-        assert isinstance(value, basestring) or value in (None, False)
+        assert isinstance(value, basestring) or value is None
         super(CharDescriptor, self).__set__(instance, value or '')
 
-    def __get__(self, instance, owner):
-        value = super(CharDescriptor, self).__get__(instance, owner)
-        if value is False:
-            value = None
-        return value
-
 
 class BinaryDescriptor(FieldDescriptor):
     default = None
 
     def __set__(self, instance, value):
-        assert (isinstance(value, (basestring, buffer))
-            or value in (None, False))
-        super(BinaryDescriptor, self).__set__(instance, value or '')
+        assert isinstance(value, (basestring, buffer)) or value is None
+        super(BinaryDescriptor, self).__set__(instance, value)
 
 
 class IntegerDescriptor(FieldDescriptor):
@@ -143,8 +151,6 @@ class NumericDescriptor(FieldDescriptor):
 class ReferenceDescriptor(FieldDescriptor):
     def __get__(self, instance, owner):
         value = super(ReferenceDescriptor, self).__get__(instance, owner)
-        if instance._parent_name == self.name:
-            value = instance._parent
         if isinstance(value, basestring):
             model_name, id = value.split(',', 1)
             if model_name:
@@ -158,7 +164,6 @@ class ReferenceDescriptor(FieldDescriptor):
         if isinstance(value, basestring):
             assert value.startswith(',')
         elif isinstance(value, Model):
-            assert value.id > 0 and not value._changed
             assert value._config == instance._config
         super(ReferenceDescriptor, self).__set__(instance, value)
 
@@ -172,25 +177,25 @@ class DateDescriptor(FieldDescriptor):
         return value
 
     def __set__(self, instance, value):
-        assert isinstance(value, datetime.date) or value in (None, False)
+        assert isinstance(value, datetime.date) or value is None
         super(DateDescriptor, self).__set__(instance, value)
 
 
 class DateTimeDescriptor(FieldDescriptor):
     def __set__(self, instance, value):
-        assert isinstance(value, datetime.datetime) or value in (None, False)
+        assert isinstance(value, datetime.datetime) or value is None
         super(DateTimeDescriptor, self).__set__(instance, value)
 
 
 class TimeDescriptor(FieldDescriptor):
     def __set__(self, instance, value):
-        assert isinstance(value, datetime.time) or value in (None, False)
+        assert isinstance(value, datetime.time) or value is None
         super(TimeDescriptor, self).__set__(instance, value)
 
 
 class DictDescriptor(FieldDescriptor):
     def __set__(self, instance, value):
-        assert isinstance(value, dict) or value in (None, False)
+        assert isinstance(value, dict) or value is None
         super(DictDescriptor, self).__set__(instance, value)
 
 
@@ -198,12 +203,8 @@ class Many2OneDescriptor(FieldDescriptor):
     def __get__(self, instance, owner):
         relation = Model.get(self.definition['relation'], instance._config)
         value = super(Many2OneDescriptor, self).__get__(instance, owner)
-        if instance._parent_name == self.name:
-            value = instance._parent
-        if isinstance(value, (int, long)) and value is not False:
+        if isinstance(value, (int, long)):
             value = relation(value)
-        elif not value:
-            value = None
         if self.name in instance._values:
             instance._values[self.name] = value
         return value
@@ -211,7 +212,6 @@ class Many2OneDescriptor(FieldDescriptor):
     def __set__(self, instance, value):
         assert isinstance(value, (Model, NoneType))
         if value:
-            assert value.id > 0 and not value._changed
             assert value._config == instance._config
         super(Many2OneDescriptor, self).__set__(instance, value)
 
@@ -342,7 +342,14 @@ class One2OneEvalDescriptor(Many2OneEvalDescriptor):
 
 class One2ManyEvalDescriptor(EvalDescriptor):
     def __get__(self, instance, owner):
-        return [x.id for x in getattr(instance, self.name)]
+        # Directly use _values to prevent infinite recursion with
+        # One2ManyDescriptor which could evaluate this field to decode the
+        # context
+        value = instance._values.get(self.name, [])
+        if isinstance(value, ModelList):
+            return [x.id for x in value]
+        else:
+            return value
 
 
 class Many2ManyEvalDescriptor(One2ManyEvalDescriptor):
@@ -411,7 +418,8 @@ class MetaModelFactory(object):
                 for field_name, definition in dict['_fields'].iteritems():
                     if field_name == 'id':
                         continue
-                    Descriptor = self.descriptors[definition['type']]
+                    Descriptor = self.descriptors.get(definition['type'],
+                        FieldDescriptor)
                     dict[field_name] = Descriptor(field_name, definition)
                     VDescriptor = self.value_descriptors.get(
                             definition['type'], ValueDescriptor)
@@ -469,27 +477,9 @@ class ModelList(list):
         ctx.update(decoder.decode(self.context) if self.context else {})
         return ctx
 
-    def append(self, record):
-        assert isinstance(record, Model)
-        if self.parent:
-            assert record._config == self.parent._config
-        elif self:
-            assert record._config == self[0]._config
-        assert record._parent is None
-        assert not record._parent_field_name
-        assert not record._parent_name
-        record._parent = self.parent
-        record._parent_field_name = self.parent_field_name
-        record._parent_name = self.parent_name
-        res = super(ModelList, self).append(record)
-        self._changed()
-        return res
-    append.__doc__ = list.append.__doc__
-
-    def extend(self, iterable):
-        iterable = list(iterable)
+    def __check(self, records):
         config = None
-        for record in iterable:
+        for record in records:
             assert isinstance(record, Model)
             if self.parent:
                 assert record._config == self.parent._config
@@ -499,13 +489,30 @@ class ModelList(list):
                 assert record._config == config
             else:
                 config = record._config
-        for record in iterable:
+        for record in records:
             assert record._parent is None
             assert not record._parent_field_name
             assert not record._parent_name
             record._parent = self.parent
             record._parent_field_name = self.parent_field_name
             record._parent_name = self.parent_name
+
+            # Set parent field to trigger on_change
+            if self.parent and self.parent_name in record._fields:
+                definition = record._fields[self.parent_name]
+                if definition['type'] in ('many2one', 'reference'):
+                    setattr(record, self.parent_name, self.parent)
+
+    def append(self, record):
+        self.__check([record])
+        res = super(ModelList, self).append(record)
+        self._changed()
+        return res
+    append.__doc__ = list.append.__doc__
+
+    def extend(self, iterable):
+        iterable = list(iterable)
+        self.__check(iterable)
         res = super(ModelList, self).extend(iterable)
         self._changed()
         return res
@@ -599,11 +606,9 @@ class Model(object):
                 getattr(self, field_name).extend(value)
             else:
                 if definition['type'] == 'many2one':
-                    if (isinstance(value, (int, long)) and value is not False):
+                    if isinstance(value, (int, long)):
                         relation = Model.get(definition['relation'])
                         value = relation(value)
-                    elif value is False:
-                        value = None
                 setattr(self, field_name, value)
     __init__.__doc__ = object.__init__.__doc__
 
@@ -692,6 +697,13 @@ class Model(object):
         self.reload()
         return True
 
+    @classmethod
+    def duplicate(cls, records, default=None):
+        'Duplicate the record'
+        ids = cls._proxy.copy([r.id for r in records], default,
+            cls._config.context)
+        return [cls(id) for id in ids]
+
     def click(self, button):
         'Click on button'
         self.save()
@@ -749,6 +761,7 @@ class Model(object):
             self._config.context))
 
     def _default_set(self, values):
+        fieldnames = []
         for field, value in values.iteritems():
             if '.' in field:
                 continue
@@ -767,6 +780,9 @@ class Model(object):
                     field)
             else:
                 self._values[field] = value
+            fieldnames.append(field)
+        for field in sorted(fieldnames):
+            self._on_change(field)
 
     def _get_eval(self):
         values = dict((x, getattr(self, '__%s_eval' % x))
@@ -774,23 +790,39 @@ class Model(object):
         values['id'] = self.id
         return values
 
-    def _get_on_change_value(self):
+    def _get_on_change_values(self, skip=None, fields=None):
         values = {'id': self.id}
-        for field, definition in self._fields.iteritems():
-            if field in self._values and field != 'id':
-                if definition['type'] == 'one2many':
-                    values[field] = [x._get_on_change_value()
-                        for x in getattr(self, field)]
-                else:
-                    values[field] = getattr(self, '__%s_eval' % field)
+        if fields:
+            definitions = ((f, self._fields[f]) for f in fields)
+        else:
+            definitions = self._fields.iteritems()
+        for field, definition in definitions:
+            if not fields:
+                if field == 'id' or (skip and field in skip):
+                    continue
+                if (self.id >= 0
+                        and (field not in self._values
+                            or field not in self._changed)):
+                    continue
+            if definition['type'] == 'one2many':
+                values[field] = [x._get_on_change_values(
+                        skip={definition.get('relation_field', '')})
+                    for x in getattr(self, field)]
+            elif (definition['type'] in ('many2one', 'reference')
+                    and self._parent_name == definition['name']
+                    and self._parent):
+                values[field] = self._parent._get_on_change_values(
+                    skip={self._parent_field_name})
+                if definition['type'] == 'reference':
+                    values[field] = (
+                        self._parent.__class__.__name__, values[field])
+            else:
+                values[field] = getattr(self, '__%s_eval' % field)
         return values
 
     def _on_change_args(self, args):
         res = {}
-        values = self._get_on_change_value()
-        if self._parent:
-            values['_parent_%s' % self._parent_name] = \
-                _EvalEnvironment(self._parent)
+        values = _EvalEnvironment(self, 'on_change')
         for arg in args:
             scope = values
             for i in arg.split('.'):
@@ -942,7 +974,7 @@ class Wizard(object):
                 # Filter only modified values
                 data = {self.form_state:
                     dict((k, v) for k, v in
-                        self.form._get_on_change_value().iteritems()
+                        self.form._get_on_change_values().iteritems()
                         if k in self.form._values)}
             else:
                 data = {}
@@ -965,3 +997,17 @@ class Wizard(object):
 
         if self.state == self.end_state:
             self._proxy.delete(self.session_id, self._config.context)
+
+
+class Report(object):
+    'Report class for Tryton reports'
+
+    def __init__(self, name, config=None, context=None):
+        super(Report, self).__init__()
+        self.name = name
+        self._config = config or proteus.config.get_config()
+        self._context = context or {}
+        self._proxy = self._config.get_proxy(name, type='report')
+
+    def execute(self, models, data):
+        return self._proxy.execute([m.id for m in models], data, self._context)
